/*
 * Simulator for fault-tolerance
 * For use with Harness/FT-MPI
 * A number of random processes fail during computation
 * and the processes are restarted
 * This is the FT-MPI version - fault-tolerance is implemented
 * Author: Knut Imar Hagen
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include "mpi.h"

int totalsize = 2000; //total size of the square matrix, length & width
MPI_Comm comm;

void help(int rank) {
	if (rank == 0) {
		printf("usage: simulator <procfail> <numloops>\n");
		printf("    procfail    - Number of randomly failing ");
		printf("processes during the computation\n");
		printf("    numloops    - Number of loops of computation, ");
		printf("to adjust computation time\n");
		printf("    NOTE: procfail must be less or equal to numloops\n");
	}
	MPI_Finalize();
	exit(0);
}

int main(int argc, char **argv) {
	int rank,p,rc,rs,procfail,numloop,*failingprocs;
	double *submat;
	comm = MPI_COMM_WORLD;

	rc = MPI_Init(&argc, &argv);
	if (rc==MPI_INIT_RESTARTED_NODE) {
		printf("MPI_Init says I am a restarted process\n"); fflush(stdout);
		rs = 1;
	}
	else {
		printf("MPI_Init [%d]\n", rc); fflush(stdout);
		rs = 0;
	}

	rc = MPI_Comm_rank(comm, &rank);
	rc = MPI_Comm_size(comm, &p);

	//If there are not 2 arguments, print help and quit
	if (argc != 3)
		help(rank);

	procfail = atoi(argv[1]);
	numloop = atoi(argv[2]);

	//number of failing processes must be less or equal to number of loops
	if (procfail > numloop || procfail < 0 || numloop < 0) {
		help(rank);
	}

	/* If the simulator shall fail some processes, let rank 0 make a random
 	 * list of the failing processes and send the list to every process
 	 */
	if (procfail > 0 && rs == 0) {
		failingprocs = malloc(sizeof(int) * procfail);
		memset(failingprocs,-1,sizeof(int) * procfail);
		if (rank == 0 && rs == 0) {
			srand(time(NULL));
			int k;
			for (k=0;k<procfail;k++) {
				int tmp = 0;
				int ok = 1;
				do {
					ok = 1;
					tmp = rand()%p;
					int l;
					for (l=0;l<k;l++) {
						if (failingprocs[l] == tmp)
							ok = 0;
					}
				}while(!ok);
				failingprocs[k] = tmp;
			}
		}
		MPI_Bcast(failingprocs,procfail,MPI_INT,0,comm);
	}

	//Allocate the submatrix
	submat = malloc(sizeof(double) * totalsize * totalsize / p);
	memset(submat, 0, sizeof(double) * totalsize * totalsize / p);

	/* Initialize the submatrix
 	 * the first element will always be rank number + 1
 	 */
	int i,j;
	for (i=0;i<totalsize/p;i++) {
		for (j=0;j<totalsize;j++) {
			submat[i*totalsize + j] = (rank+1)*(j+i+1);
		}
	}

	/* If the number of loops is set higher than 0, the computation will run
	 * the given number of times, and the processes listed to be failed
	 * will fail in a sequentially order in a regular manner
	 */
	if (numloop > 0 ) {
		int countfail = 0;
		int upcountfail = 0;
		if (procfail == 0)
			procfail = 1;
		else
			upcountfail = numloop/procfail;
		for (i=0;i<numloop;i++) {
			if (numloop/procfail == upcountfail && rs == 0) {
				if (countfail < procfail
						&& failingprocs[countfail++] == rank) {
					printf("I am process %d and I am failing\n",rank);
					fflush(stdout);
					exit(15);
				}
				upcountfail = 0;
			}
			upcountfail++;
			for (j=0;j<(totalsize*totalsize/p)/2;j++) {
				double swp = submat[j];
				submat[j] = submat[totalsize*totalsize/p-j];
				submat[totalsize*totalsize/p-j] = swp;
			}
			for (j=0;j<(totalsize*totalsize/p)/2;j++) {
				double swp = submat[totalsize*totalsize/p-j];
				submat[totalsize*totalsize/p-j] = submat[j];
				submat[j] = swp;
			}
		}
	}

	do {
		rc = MPI_Barrier(comm);
		if (rc == MPI_ERR_OTHER) {
			MPI_Comm newcomm;
			newcomm = FT_MPI_CHECK_RECOVER;
			MPI_Comm_dup(comm,&newcomm);
			MPI_Comm_free(&comm);
			comm = newcomm;
		}
	}while(rc == MPI_ERR_OTHER);

	//Processes that have not failed will say that they are alive
	printf("Process #%d of %d: submat[0]=%f\n", rank,p,submat[0]);
	fflush(stdout);
	double sum = 0;
	//The first element in the matrix will be reduced
	rc = MPI_Reduce(submat,&sum,1,MPI_DOUBLE,MPI_SUM,0,comm);
	if (rank==0) {
		printf("The result from the MPI_Reduce() operation is %f\n",sum);
		fflush(stdout);
	}
	MPI_Finalize();
}
